# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import time

from collections import namedtuple, defaultdict, OrderedDict
from functools import partial
from .structures import Single, Call, Group, Mixed, TreeElement, prepare_cache
from .structures import Graph
from ..utils import OrderedCounter
from ..persistence import persistence


Edge = namedtuple("Edge", "node count")


class TreeVisitor(object):

    def __init__(self):
        self.nodes = []
        self.edges = []
        self.delegated = {
            'initial': Edge(0, 1)
        }
        self.nid = 0
        self.min_duration = defaultdict(partial(int, 1000^10))
        self.max_duration = defaultdict(partial(int, 0))
        self.keep = None

    def update_durations(self, duration, tid):
        self.max_duration[tid] = max(self.max_duration[tid], duration)
        self.min_duration[tid] = min(self.min_duration[tid], duration)

    def update_node(self, node):
        node['mean'] = node['duration'] / node['count']
        node['info'].update_by_node(node)
        node['info'] = repr(node['info'])

    def to_dict(self):
        for node in self.nodes:
            n = node['node']
            self.update_node(n)
            self.update_durations(n['duration'], n['trial_id'])

        self.update_edges()
        return {
            'nodes': self.nodes,
            'edges': self.edges,
            'min_duration': self.min_duration,
            'max_duration': self.max_duration
        }

    def update_edges(self):
        pass

    def add_node(self, node):
        self.nodes.append(node.to_dict(self.nid))
        original = self.nid
        self.nid += 1
        return original

    def add_edge(self, source, target, count, typ):
        self.edges.append({
            'source': source,
            'target': target,
            'count': count,
            'type': typ
        })

    def visit_call(self, call):
        caller_id = self.add_node(call.caller)
        self.nodes[caller_id]['repr'] = repr(call)
        callees = call.called.visit(self)
        pos = 1
        for callee_id in callees:
            self.add_edge(caller_id, callee_id, pos, 'call')
            pos += 1
        return [caller_id]

    def visit_group(self, group):
        result = []
        for element in group.nodes.values():
            result += element.visit(self)
        return result

    def visit_single(self, single):
        return [self.add_node(single)]

    def visit_mixed(self, mixed):
        mixed.mix_results()
        node_id = mixed.first.visit(self)
        self.nodes[node_id[0]]['duration'] = mixed.duration
        return node_id

    def visit_treeelement(self, tree_element):
        return []


class NoMatchVisitor(TreeVisitor):

    def update_edges(self):
        for edge in self.edges:
            if edge['type'] in ['return', 'call']:
                edge['count'] = ''

    def use_delegated(self):
        result = self.delegated
        self.delegated = {}
        return result

    def solve_delegation(self, node_id, node_count, delegated):
        self.solve_cis_delegation(node_id, node_count, delegated)
        self.solve_ret_delegation(node_id, node_count, delegated)

    def solve_cis_delegation(self, node_id, node_count, delegated):
        # call initial sequence
        for typ in ['call', 'initial', 'sequence']:
            if typ in delegated:
                edge = delegated[typ]
                self.add_edge(edge.node, node_id, node_count, typ)

    def solve_ret_delegation(self, node_id, node_count, delegated):
        if 'return' in delegated:
            edge = delegated['return']
            self.add_edge(node_id, edge.node, edge.count, 'return')

    def visit_call(self, call):
        delegated = self.use_delegated()
        caller_id = self.add_node(call.caller)
        self.nodes[caller_id]['repr'] = repr(call)

        if delegated:
            self.solve_delegation(caller_id, call.count, delegated)

        self.delegated['call'] = Edge(caller_id, 1)
        self.delegated['return'] = Edge(caller_id, 1)

        call.called.visit(self)
        return caller_id, call

    def visit_group(self, group):
        delegated = self.use_delegated()

        node_map = {}
        for element in group.nodes.values():
            node_id, node = element.visit(self)
            node_map[node] = node_id

        self.solve_cis_delegation(node_map[group.next], group.count, delegated)
        self.solve_ret_delegation(node_map[group.last], group.count, delegated)

        for previous, edges in group.edges.items():
            for next, count in edges.items():
                self.add_edge(node_map[previous], node_map[next],
                              count, 'sequence')

        return node_map[group.next], group.next

    def visit_single(self, single):
        delegated = self.use_delegated()
        node_id = self.add_node(single)
        self.nodes[node_id]['repr'] = repr(single)

        if delegated:
            self.solve_delegation(node_id, single.count, delegated)
        return node_id, single

    def visit_mixed(self, mixed):
        mixed.mix_results()
        node_id, node = mixed.first.visit(self)
        self.nodes[node_id]['duration'] = mixed.duration
        return node_id, node

    def visit_treeelement(self, tree_element):
        return None, tree_element


class ExactMatchVisitor(NoMatchVisitor):

    def visit_single(self, single):
        s = Single(single.activation)
        s.level = single.level
        s.use_id = False
        return s

    def visit_mixed(self, mixed):
        mixed.use_id = False
        m = Mixed(mixed.elements[0].visit(self))
        for element in elements[1:]:
            m.add_element(element.visit(self))
        m.level = mixed.level
        return m

    def visit_group(self, group):
        nodes = group.nodes.keys()
        g = Group()
        g.use_id = False
        g.initialize(nodes[1].visit(self),nodes[0].visit(self))
        for element in nodes[2:]:
            g.add_subelement(element.visit(self))
        g.level = group.level
        return g

    def visit_call(self, call):
        caller = call.caller.visit(self)
        called = call.called.visit(self)
        c = Call(caller, called)
        c.use_id = False
        c.level = call.level
        return c

    def visit_treeelement(self, tree_element):
        return tree_element


class NamespaceVisitor(NoMatchVisitor):

    def __init__(self):
        super(NamespaceVisitor, self).__init__()
        self.context = {}
        self.context_edges = {}
        self.namestack = []

    def update_edges(self):
        pass

    def namespace(self):
        return ' '.join(self.namestack)

    def update_namespace_node(self, node, single):
        node['count'] += single.count
        node['duration'] += single.duration
        node['info'].add_activation(single.activation)

    def add_node(self, single):
        self.namestack.append(single.name_id())
        namespace = self.namespace()
        self.namestack.pop()
        if namespace in self.context:
            context = self.context[namespace]
            self.update_namespace_node(context['node'], single)

            return self.context[namespace]['index']

        single.namespace = namespace
        result = super(NamespaceVisitor, self).add_node(single)
        self.context[namespace] = self.nodes[-1]
        return result

    def add_edge(self, source, target, count, typ):

        edge = "{} {} {}".format(source, target, typ)
        if not edge in self.context_edges:
            super(NamespaceVisitor, self).add_edge(source, target,
                                                           count, typ)
            self.context_edges[edge] = self.edges[-1]
        else:
            e = self.context_edges[edge]
            self.context_edges[edge]['count'] += count

    def visit_call(self, call):
        self.namestack.append(call.caller.name_id())
        result = super(NamespaceVisitor, self).visit_call(call)
        self.namestack.pop()
        return result

    def visit_mixed(self, mixed):
        node_id, node = None, None
        for element in mixed.elements:
            node_id, node = element.visit(self)
        return node_id, node


def sequence(previous, next):
    if isinstance(next, Group):
        next.add_subelement(previous)
        return next
    return Group().initialize(previous, next)


def list_to_call(stack):
    group = stack.pop()
    next = group.pop()
    while group:
        previous = group.pop()
        next = sequence(previous, next)
    caller = stack[-1].pop()
    call = Call(caller, next)
    call.level = caller.level
    next.level = caller.level + 1
    stack[-1].append(call)


def generate_graph(trial):
    """ Returns activation graph """
    activations = [Single(act) for act in trial.activations()]
    if not activations:
        t = TreeElement(level=0)
        t.trial_id = trial.id
        return t

    current = activations[0]
    stack = [[current]]
    level = OrderedDict()
    current.level = level[current.id] = 0

    for i in range(1, len(activations)):
        act = activations[i]
        act.level = level[act.id] = level[act.parent] + 1
        last = stack[-1][-1]
        if act.level == last.level:
            # act in the same level, add act to sequence
            stack[-1].append(act)
        elif act.level > last.level:
            # last called act
            stack.append([act])
        else:
            # act is in higher level than last
            # create a call for last group
            # add act to existing sequence
            while last.level > act.level:
                list_to_call(stack)
                last = stack[-1][-1]
            stack[-1].append(act)

    while len(stack) > 1:
        list_to_call(stack)

    return(stack[-1][-1])


cache = prepare_cache(
    lambda self, *args, **kwargs: 'trial {}'.format(self.trial_id))


class TrialGraph(Graph):

    def __init__(self, trial_id, use_cache=True,
                 width=500, height=500, mode=3):
        self._graph = None
        self.trial_id = trial_id
        self.use_cache = use_cache

        self.width = width
        self.height = height
        self.mode = mode
        self._modes = {
            0: self.tree,
            1: self.no_match,
            2: self.exact_match,
            3: self.namespace_match
        }

    @cache('graph')
    def graph(self, trial=None):
        """Generate an activation tree structure"""
        if not trial:
            from ..models import Trial
            trial = Trial(self.trial_id)
        if self._graph == None:
            self._graph = generate_graph(trial)
        return self._graph

    @cache('tree')
    def tree(self, trial=None):
        """Convert tree structure into dict tree structure"""
        graph = self.graph(trial)
        visitor = TreeVisitor()
        graph.visit(visitor)
        return visitor.to_dict()

    @cache('no_match')
    def no_match(self, trial=None):
        """Convert tree structure into dict graph without node matchings"""
        graph = self.graph(trial)
        visitor = NoMatchVisitor()
        graph.visit(visitor)
        return visitor.to_dict()

    @cache('exact_match')
    def exact_match(self, trial=None):
        """Convert tree structure into dict graph and match equal calls"""
        graph = self.graph(trial)
        graph = graph.visit(ExactMatchVisitor())
        visitor = NoMatchVisitor()
        graph.visit(visitor)
        return visitor.to_dict()

    @cache('namespace_match')
    def namespace_match(self, trial=None):
        """Convert tree structure into dict graph and match namespaces"""
        graph = self.graph(trial)
        visitor = NamespaceVisitor()
        graph.visit(visitor)
        return visitor.to_dict()

    def _repr_html_(self, trial=None):
        """ Display d3 graph on ipython notebook """
        uid = str(int(time.time()*1000000))

        result = """
            <div class="nowip-trial" data-width="{width}"
                 data-height="{height}" data-uid="{uid}"
                 data-id="{id}">
                {data}
            </div>
        """.format(
            uid=uid, id=self.trial_id,
            data=self.escape_json(self._modes[self.mode](trial)),
            width=self.width, height=self.height)
        return result



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import itertools
import time

from copy import deepcopy, copy
from collections import namedtuple, defaultdict, OrderedDict
from ..utils import OrderedCounter, concat_iter, hashabledict
from .structures import prepare_cache, Graph


def fix_caller_id(graph):
    called = {}
    seq = defaultdict(list)
    nodes = graph['nodes']
    edges = [hashabledict(x) for x in graph['edges']]
    for edge in edges:
        if edge['type'] == 'call':
            called[edge['target']] = edge['source']
        if edge['type'] == 'sequence':
            seq[edge['source']].append(edge)

    visited = set()
    while called:
        t = {}
        for nid, parent in called.items():
            nodes[nid]['caller_id'] = parent
            visited.add(nid)
            for e in seq[nid]:
                if not e['target'] in visited:
                    t[e['target']] = parent
        called = t


def prepare_graph(g1, g2):
    """Fix ids on graphs, match the <main> node
    Return hashable nodes1 and nodes2 from g1 and g2, respectively"""
    fix_caller_id(g1)
    fix_caller_id(g2)
    nodes1 = [hashabledict(x) for x in g1['nodes']]
    nodes2 = [hashabledict(x) for x in g2['nodes']]
    g1['hnodes'], g2['hnodes'] = nodes1, nodes2
    g1['node_indexes'] = set(range(len(nodes1)))
    g2['node_indexes'] = set(range(len(nodes2)))
    g1['levels'] = defaultdict(set)
    g2['levels'] = defaultdict(set)
    for n in nodes1:
        g1['levels'][n['node']['level']].add(n['index'])
    for n in nodes2:
        g2['levels'][n['node']['level']].add(n['index'])
    if nodes1 and nodes2:
        nodes1[0]['name'] = '<main>'
        nodes2[0]['name'] = '<main>'
        g1['max_level'] = max(g1['levels'].keys())
        g2['max_level'] = max(g2['levels'].keys())
    else:
        g1['max_level'] = -1
        g2['max_level'] = -1

    return nodes1, nodes2


def lcs(a, b, eq=lambda x, y: x == y):
    lengths = [[0 for j in range(len(b)+1)] for i in range(len(a)+1)]
    # row 0 and column 0 are initialized to 0 already
    for i, x in enumerate(a):
        for j, y in enumerate(b):
            if eq(x, y):
                lengths[i+1][j+1] = lengths[i][j] + 1
            else:
                lengths[i+1][j+1] = \
                    max(lengths[i+1][j], lengths[i][j+1])
    # read the substring out from the matrix
    result_a, result_b = OrderedCounter(), OrderedCounter()
    x, y = len(a), len(b)
    while x != 0 and y != 0:
        if lengths[x][y] == lengths[x-1][y]:
            x -= 1
        elif lengths[x][y] == lengths[x][y-1]:
            y -= 1
        else:
            result_a[a[x-1]] = b[y-1]
            result_b[b[y-1]] = a[x-1]
            x -= 1
            y -= 1
    return result_a, result_b


def cmp_node_fn(nodes1, nodes2):
    def cmp_node(x, y, check_caller=True):
        if x['name'] != y['name']:
            return False
        if check_caller:
            if x['caller_id'] is None and y['caller_id'] is not None:
                return False
            if x['caller_id'] is not None and y['caller_id'] is None:
                return False
            if x['caller_id'] is None and y['caller_id'] is None:
                return True
            caller1, caller2 = nodes1[x['caller_id']], nodes2[y['caller_id']]
            return caller1['name'] == caller2['name']
        return True
    return cmp_node


def cmp_edge_fn(nodes1, nodes2):
    cmp_node = cmp_node_fn(nodes1, nodes2)
    def cmp_edge(x, y):
        if x['type'] != y['type']:
            return False
        if not cmp_node(nodes1[x['source']], nodes2[y['source']]):
            return False
        if not cmp_node(nodes1[x['target']], nodes2[y['target']]):
            return False
        return True
    return cmp_edge


def first_solution(nodes1, nodes2):
    "Use LCS and return an initial mapping between nodes"
    mapping, _ = lcs(nodes1, nodes2, cmp_node_fn(nodes1, nodes2))
    return mapping


class MappingToGraph(object):

    def __init__(self, g1, g2, mapping):
        self.id = 0
        self.nodes = []
        self.edges = []
        self.context_edges = {}
        self.old_to_new = {}
        self.max_duration = dict(concat_iter(
            g1['max_duration'].items(), g2['max_duration'].items()))

        self.min_duration = dict(concat_iter(
            g1['min_duration'].items(), g2['min_duration'].items()))

        self.g1, self.g2 = g1, g2
        self.nodes1, self.nodes2 = g1['hnodes'], g2['hnodes']
        del g1['hnodes'], g2['hnodes']
        del g1['node_indexes'], g2['node_indexes']
        del g1['levels'], g2['levels']
        del g1['max_level'], g2['max_level']
        self.mapping = mapping
        self.merge()

    def merge(self):
        for a, b in self.mapping.items():
            n = deepcopy(a)
            del n['node']
            n['node1'] = a['node']
            n['node2'] = b['node']
            n['node1']['original'] = a['index']
            n['node2']['original'] = b['index']
            n['index'] = self.id
            a['node']['diff'] = self.id
            b['node']['diff'] = self.id
            self.old_to_new[(1, a['index'])] = self.id
            self.old_to_new[(2, b['index'])] = self.id
            self.id += 1
            self.nodes.append(n)

        self.add_nodes(self.nodes1, 1)
        self.add_nodes(self.nodes2, 2)
        self.add_edges(self.g1['edges'], 1)
        self.add_edges(self.g2['edges'], 2)

    def add_nodes(self, nodes, ng):
        for node in nodes:
            if not (ng, node['index']) in self.old_to_new:
                nid = self.add_node(node)
                self.old_to_new[(ng, node['index'])] = nid
                node['node']['diff'] = nid

    def add_edges(self, edges, ng):
        for edge in edges:
            if (ng, edge['source']) in self.old_to_new:
                self.add_edge(self.old_to_new[(ng, edge['source'])],
                              self.old_to_new[(ng, edge['target'])],
                              edge, ng)

    def add_node(self, node):
        n = deepcopy(node)
        node_id = n['index'] = self.id
        n['node']['original'] = node['index']
        self.nodes.append(n)
        self.id += 1
        return node_id

    def add_edge(self, source, target, edge, ng):
        edge_key = "{} {} {}".format(source, target, edge['type'])

        if not edge_key in self.context_edges:
            e = deepcopy(edge)
            e['source'] = source
            e['target'] = target
            e['trial'] = ng
            self.edges.append(e)
            self.context_edges[edge_key] = e
        else:
            e = self.context_edges[edge_key]
            e['count'] = (e['count'], edge['count'])
            e['trial'] = 0

    def to_dict(self):
        return {
            'max_duration': self.max_duration,
            'min_duration': self.min_duration,
            'nodes': self.nodes,
            'edges': self.edges,
        }


class Similarity(object):

    def __init__(self, g1, g2):
        self.g1, self.g2 = g1, g2
        self.nodes1, self.nodes2 = g1['hnodes'], g2['hnodes']
        self.total_nodes = float(self.count_union_nodes()) or 1
        self.total_edges = float(self.count_union_edges()) or 1
        self.edges1, self.edges2 = defaultdict(set), defaultdict(set)
        for e in g1['edges']:
            self.edges1[e['source']].add((e['target'], e['type']))
        for e in g2['edges']:
            self.edges2[e['source']].add((e['target'], e['type']))

    def count_union_nodes(self):
        nodes1, nodes2 = self.nodes1, self.nodes2
        if len(nodes2) < len(nodes1):
            nodes2, nodes1 = nodes1, nodes2
        cmp_node = cmp_node_fn(nodes1, nodes2)

        used = [False] * len(nodes1)
        for node2 in nodes2:
            for i, node1 in enumerate(nodes1):
                if not used[i] and cmp_node(node1, node2):
                    used[i] = node2
                    break
        return sum(1 for i, _ in enumerate(used) if not used[i]) + len(nodes2)

    def count_union_edges(self):
        nodes1, nodes2 = self.nodes1, self.nodes2
        edges1, edges2 = self.g1['edges'], self.g2['edges']
        if len(edges2) < len(edges1):
            edges2, edges1 = edges1, edges2
            nodes2, nodes1 = nodes1, nodes2
        cmp_edge = cmp_edge_fn(nodes1, nodes2)

        used = [False] * len(edges1)
        for edge2 in edges2:
            for i, edge1 in enumerate(edges1):
                if not used[i] and cmp_edge(edge1, edge2):
                    used[i] = edge2
                    break
        return sum(1 for i, _ in enumerate(used) if not used[i]) + len(edges2)

    def edge_intersection(self, mapping):
        edges = 0
        for node1, node2 in mapping.items():
            targets1 = self.edges1[node1['index']]
            targets2 = self.edges2[node2['index']]
            for target, typ in targets1:
                target1 = self.nodes1[target]
                if not target1 in mapping:
                    continue
                if (mapping[target1]['index'], typ) in targets2:
                    edges += 1
        return float(edges)

    def __call__(self, mapping):
        node_sim = (float(len(mapping)) / self.total_nodes)
        edge_sim = self.edge_intersection(mapping) / self.total_edges
        result = node_sim + edge_sim
        return result


def greedy(g1, g2):
    nodes1, nodes2 = prepare_graph(g1, g2)
    mapping = first_solution(nodes1, nodes2)
    return MappingToGraph(g1, g2, mapping).to_dict()


def neighborhood(k, g1, g2, mapping):
    nodes1, nodes2 = g1['hnodes'], g2['hnodes']
    cmp_node = cmp_node_fn(nodes1, nodes2)
    tried = set()
    used1 = {n['index'] for n in mapping.keys()}
    used2 = {n['index'] for n in mapping.values()}
    reverse = {n2:n1 for n1, n2 in mapping.items()}
    max_level = min(g1['max_level'], g2['max_level'])

    def add_to_mapping(to_add, new_mapping, swapped):
        added = []
        for node_id1, node_id2 in to_add:
            if swapped:
                node_id1, node_id2 = node_id2, node_id1
            node1, node2 = nodes1[node_id1], nodes2[node_id2]
            if cmp_node(node1, node2):
                added.append((node_id1, node_id2))
                new_mapping[node1] = node2
        return tuple(added)

    def permutate(group1, group2, allow_different_caller):
        for n1, n2 in itertools.product(group1, group2):
            node1, node2 = nodes1[n1], nodes2[n2]
            if id(mapping[node1]) == id(node2):
                # already on mapping
                continue
            if not cmp_node(node1, node2, not allow_different_caller):
                # different nodes
                continue
            new_mapping = copy(mapping)
            if node2 in reverse:
                del new_mapping[reverse[node2]]
            new_mapping[node1] = node2
            for m in neighborhood(1, g1, g2, new_mapping):
                yield m

    if k == 1:
        not_mapped1 = g1['node_indexes'] - used1
        not_mapped2 = g2['node_indexes'] - used2
        swapped = False
        if len(not_mapped2) > len(not_mapped1):
            swapped = True
            not_mapped1, not_mapped2 = not_mapped2, not_mapped1

        possibilities = [zip(x, not_mapped2) for x in itertools.permutations(
            not_mapped1, len(not_mapped2))]
        result = []
        for full_map in possibilities:
            for i in range(1, len(not_mapped2) + 1):
                for to_add in itertools.combinations(full_map, i):
                    if to_add in tried:
                        continue
                    new_mapping = copy(mapping)
                    to_add = add_to_mapping(to_add, new_mapping, swapped)
                    if to_add in tried:
                        continue
                    tried.add(to_add)
                    yield new_mapping

    if k == 2:
        for i in range(1, max_level + 1):
            for m in permutate(g1['levels'][i], g2['levels'][i], False):
                yield m

    if k == 3:
        for i in range(1, max_level + 1):
            for m in permutate(g1['node_indexes'], g2['node_indexes'], True):
                yield m

    yield mapping


def vnd(g1, g2, neighborhoods=3, time_limit=0):
    if time_limit == 0:
        time_limit = 315569000
    time_limit *= 1000
    nodes1, nodes2 = prepare_graph(g1, g2)
    sim = Similarity(g1, g2)
    start = time.time()
    f = first_solution(nodes1, nodes2)
    best = (f, sim(f))
    k = 1
    while k <= neighborhoods and (time.time() - start) < time_limit:
        try:
            current = max(((n, sim(n)) for n in neighborhood(k, g1, g2, best[0])
                           if (time.time() - start) < time_limit),
                          key=lambda x: x[1])
        except ValueError:
            current = best
        if current[1] > best[1]:
            best = current
            k = 1
        else:
            k += 1
    return MappingToGraph(g1, g2, best[0]).to_dict()


def show_mapping(mapping):
    return [(x['index'], y['index'], x['name']) for x,y in mapping.items()]


def trial_mirror(fn):
    name = fn.__name__
    def calculate(self, diff=None, **kwargs):
        if not diff:
            from ..models import Diff
            diff = Diff(self.trial1_id, self.trial2_id)
        g1 = getattr(diff.trial1.graph, name)(diff.trial1)
        g2 = getattr(diff.trial2.graph, name)(diff.trial2)
        return fn(self, diff, g1, g2, **kwargs)
    return calculate


def class_param(params):
    params = params.split()
    def dec(fn):
        def calculate(self, *args, **kwargs):
            for param in params:
                value = getattr(self, param)
                if value is not None and not param in kwargs:
                    kwargs[param] = value
            return fn(self, *args, **kwargs)
        return calculate
    return dec


cache = prepare_cache(
    lambda self, *args, **kwargs: 'diff {}:{}'.format(self.trial1_id,
                                                      self.trial2_id))


class DiffGraph(Graph):

    def __init__(self, trial1_id, trial2_id, use_cache=True,
                 width=500, height=500, mode=3, view=0):
        self.trial1_id = trial1_id
        self.trial2_id = trial2_id
        self.use_cache = use_cache

        self.width = width
        self.height = height
        self.mode = mode
        self.view = view

        self._modes = {
            0: self.tree,
            1: self.no_match,
            2: self.exact_match,
            3: self.namespace_match
        }

        self._views = {
            0: "combined",
            1: "side by side",
            2: "both",
        }

        self.neighborhoods = None
        self.time_limit = None


    @class_param('neighborhoods time_limit')
    @cache('tree', 'neighborhoods time_limit')
    @trial_mirror
    def tree(self, diff, g1, g2, **kwargs):
        """Compare tree structures"""
        return vnd(g1, g2, **kwargs), g1, g2

    @class_param('neighborhoods time_limit')
    @cache('no_match', 'neighborhoods time_limit')
    @trial_mirror
    def no_match(self, diff, g1, g2, **kwargs):
        """Compare graphs without matches"""
        return vnd(g1, g2, **kwargs), g1, g2

    @class_param('neighborhoods time_limit')
    @cache('exact_match', 'neighborhoods time_limit')
    @trial_mirror
    def exact_match(self, diff, g1, g2, **kwargs):
        """Compare graphs with call matches"""
        return vnd(g1, g2, **kwargs), g1, g2

    @class_param('neighborhoods time_limit')
    @cache('combine', 'neighborhoods time_limit')
    @trial_mirror
    def namespace_match(self, diff, g1, g2, **kwargs):
        """Compare graphs with namespaces"""
        return vnd(g1, g2, **kwargs), g1, g2

    def _repr_html_(self, diff=None):
        """ Display d3 graph on ipython notebook """
        uid = str(int(time.time()*1000000))

        result = """
            <div class="nowip-diff" data-width="{width}"
                 data-height="{height}" data-uid="{uid}"
                 data-id1="{id1}" data-id2="{id2}" data-mode="{view}">
                {data}
            </div>
        """.format(
            uid=uid, id1=self.trial1_id, id2=self.trial2_id,
            view=self.view,
            data=self.escape_json(self._modes[self.mode](diff)),
            width=self.width, height=self.height)
        return result



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


import json
import time
from datetime import datetime
from collections import OrderedDict
from ..utils import calculate_duration, FORMAT, OrderedCounter, print_msg
from ..persistence import row_to_dict, persistence


try:
   import cPickle as pickle
except:
   import pickle


class Graph(object):
    def escape_json(self, data):
        data = json.dumps(data)
        return (data.replace('&', '\\u0026')
                    .replace('<', '\\u003c')
                    .replace('>', '\\u003e'))


def prepare_cache(get_type):
    def cache(name, attrs=""):
        def dec(fn):
            def load(self, *args, **kwargs):
                typ = get_type(self, *args, **kwargs)
                attributes = ' '.join(
                    [str(kwargs[a]) for a in attrs.split() if a in kwargs])
                conditions = {
                    'type': '"{}"'.format(typ),
                    'name': '"{}"'.format(name),
                    'attributes': '"{}"'.format(attributes),
                }
                if self.use_cache:
                    try:
                        for c in persistence.load('graph_cache', **conditions):
                            return pickle.loads(
                                persistence.get(c[b'content_hash']))
                    except:
                        print_msg("Couldn't load graph cache", True)
                start = time.time()
                graph = fn(self, *args, **kwargs)
                duration = time.time() - start
                try:
                    persistence.delete('graph_cache', **conditions)
                    persistence.insert('graph_cache', {
                        'type': typ,
                        'name': name,
                        'duration': duration,
                        'attributes': attributes,
                        'content_hash': persistence.put(pickle.dumps(graph)),
                    })
                except:
                    print_msg("Couldn't store graph cache", True)
                return graph
            return load
        return dec
    return cache


class TreeElement(object):

    def __init__(self, level=-1, use_id=True):
        self.duration = 0
        self.count = 1
        self.repr = ""
        self.level = level
        self.use_id = use_id
        self.trial_id = 0

    def mean(self):
        if isinstance(self.duration, tuple):
            return (self.a.duration / self.a.count,
                    self.b.duration / self.b.count)
        return self.duration / self.count

    def visit(self, visitor):
        name = "visit_{}".format(self.__class__.__name__.lower())
        return getattr(visitor, name)(self)

    def calculate_repr(self):
        pass

    def mix(self, other):
        pass

    def __hash__(self):
        if self.use_id:
            return id(self)
        return hash(self.__repr__())

    def __repr__(self):
        return self.repr

    def to_dict(self, nid):
        return {}


class Single(TreeElement):

    def __init__(self, activation):
        self.activation = activation
        self.activations = {activation}
        self.parent = activation['caller_id']
        self.id = activation['id']
        self.line = activation['line']
        self.name = activation['name']
        self.trial_id = activation['trial_id']
        self.repr = "S({0}-{1})".format(self.line, self.name)
        self.use_id = True
        self.level = 0

    @property
    def count(self):
        return sum(1 for a in self.activations)

    @count.setter
    def count(self, value):
        pass

    @property
    def duration(self):
        return sum(calculate_duration(a) for a in self.activations
                   if a['finish'] and a['start'])

    @duration.setter
    def duration(self, value):
        pass

    def mix(self, other):
        self.count += other.count
        self.duration += other.duration
        self.activations = self.activations.union(other.activations)

    def __eq__(self, other):
        return all(fn(self) == fn(other) for fn in
                   [type, lambda x: x.line, lambda x: x.name])

    def name_id(self):
        return "{0} {1}".format(self.line, self.name)

    def visit(self, visitor):
        return visitor.visit_single(self)

    def to_dict(self, nid):
        return {
            'index': nid,
            'caller_id': self.parent,
            'name': self.name,
            'node': {
                'trial_id': self.trial_id,
                'line': self.line,
                'count': self.count,
                'duration': self.duration,
                'level': self.level,
                'info': Info(self),
            }
        }

    def __hash__(self):
        return super(Single, self).__hash__()


class Mixed(TreeElement):

    def __init__(self, activation):
        self.duration = activation.duration
        self.elements = [activation]
        self.parent = activation.parent
        self.id = activation.id
        self.repr = activation.repr
        self.use_id = True
        self.level = 0

    @property
    def count(self):
        return sum(e.count for e in self.elements)

    @count.setter
    def count(self, value):
        pass

    @property
    def duration(self):
        return sum(e.duration for e in self.elements)

    @property
    def first(self):
        return next(iter(self.elements))

    @duration.setter
    def duration(self, value):
        pass

    def add_element(self, element):
        self.elements.append(element)

    def mix(self, other):
        self.elements += other.elements
        self.mix_results()

    def mix_results(self):
        it = iter(self.elements)
        initial = next(it)
        for element in it:
            initial.mix(element)

    def __hash__(self):
        return super(Mixed, self).__hash__()


class Group(TreeElement):

    def __init__(self):
        self.nodes = OrderedDict()
        self.edges = OrderedDict()
        self.duration = 0
        self.parent = None
        self.count = 1
        self.repr = ""
        self.use_id = True
        self.level = 0

    def initialize(self, previous, next):
        self.nodes[next] = Mixed(next)
        self.duration = next.duration
        self.next = next
        self.last = next
        self.add_subelement(previous)
        self.parent = next.parent
        return self

    def add_subelement(self, previous):
        next, self.next = self.next, previous
        if not previous in self.edges:
            self.edges[previous] = OrderedCounter()
        if not previous in self.nodes:
            self.nodes[previous] = Mixed(previous)
        else:
            self.nodes[previous].add_element(previous)
        self.edges[previous][next] += 1

    def calculate_repr(self):
        result = [
            "[{0}-{1}->{2}]".format(previous, count, next)
            for previous, edges in self.edges.items()
            for next, count in edges.items()
        ]

        self.repr = "G({0})".format(', '.join(result))

    def __eq__(self, other):
        return all(fn(self) == fn(other) for fn in
                   [type, lambda x: x.edges])

    def mix(self, other):
        for node, value in self.nodes.items():
            value.mix(other.nodes[node])

    def __hash__(self):
        return super(Group, self).__hash__()


class Call(TreeElement):

    def __init__(self, caller, called):
        self.caller = caller
        self.called = called
        self.called.calculate_repr()
        self.parent = caller.parent
        self.count = 1
        self.id = self.caller.id
        self.duration = self.caller.duration
        self.repr = 'C({0}, {1})'.format(self.caller, self.called)
        self.use_id = True
        self.level = 0

    def __eq__(self, other):
        return all(fn(self) == fn(other) for fn in
                   [type, lambda x: x.caller, lambda x: x.called])

    def mix(self, other):
        self.caller.mix(other.caller)
        self.called.mix(other.called)

    def __hash__(self):
        return super(Call, self).__hash__()


class Info(object):

    def __init__(self, single):
        self.title = ("Trial {trial}<br>"
                      "Function <b>{name}</b> called at line {line}").format(
            trial=single.trial_id, name=single.name, line=single.line)
        self.activations = set()
        self.duration = ""
        self.mean = ""
        self.extract_activations(single)

    def update_by_node(self, node):
        self.duration = self.duration_text(node['duration'], node['count'])
        self.mean = self.mean_text(node['mean'])
        self.activation_list = sorted(self.activations, key=lambda a: a[0])

    def add_activation(self, activation):
        self.activations.add(
            (datetime.strptime(activation['start'], FORMAT), activation))

    def extract_activations(self, single):
        for activation in single.activations:
            self.add_activation(activation)

    def duration_text(self, duration, count):
        return "Total duration: {} microseconds for {} activations".format(
            duration, count)

    def mean_text(self, mean):
        return "Mean: {} microseconds per activation".format(mean)

    def activation_text(self, activation):
        values = map(row_to_dict, persistence.load('object_value',
            function_activation_id=activation['id'], order='id'))
        values = [value for value in values if value['type'] == 'ARGUMENT']
        result = [
            "",
            "Activation #{id} from {start} to {finish} ({dur} microseconds)"
                .format(dur=calculate_duration(activation), **activation),
        ]
        if values:
            result.append("Arguments: {}".format(
                ", ".join("{}={}".format(value["name"], value["value"])
                    for value in values)))
        return result + [
            "Returned {}".format(activation['return'])
        ]

    def __repr__(self):
        result = [self.title, self.duration, self.mean]
        for activation in self.activation_list:
            result += self.activation_text(activation[1])

        return '<br/>'.join(result)

